#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/

import sys, os
sys.path.append("..")
sys.path.append(".")
import logging
import argparse
from keysql_client.client import KeySqlClient
from keysql_client.key_object import KeyObjectChar, KeyObjectNumber,KeyObjectComposition, KeyObjectMultiple
from keysql_client.keysql_store import KeySqlStore
from keysql_client.keysql_catalog import KeySqlCatalog
from keysql_client.keysql_catalog_statement import KeySqlDropCatalogStatement
from keysql_client.keysql_class_from_instance import KeySqlClassFromInstance

import locale
locale.setlocale(locale.LC_ALL, 'en_US.utf-8')

class Email:
    def __init__(self, email_address: str = None, email_type: str = None):
        self.email_address = email_address
        self.email_type = email_type

class Address:
    def __init__(self, country: str = None, city: str = None):
        self.country = country
        self.city = city

class Person:
    def __init__(self, first_name: str = None, last_name: str = None, yob : int = None, address: Address = None,emails  = []):
        self.first_name: str = first_name
        self.last_name = last_name
        self.yob: int = yob
        self.emails = emails
        self.address = address

def create_schema():
    catalog_name = "people"
    people = KeySqlCatalog(catalog_name)

    email_address = KeyObjectChar("email_address")
    email_type = KeyObjectChar("email_type")
    email = KeyObjectComposition("email",[email_address,email_type])
    emails = KeyObjectMultiple("emails",email)
    first_name = KeyObjectChar("first_name")
    last_name = KeyObjectChar("last_name")
    yob = KeyObjectNumber("yob")
    country = KeyObjectChar("country")
    city = KeyObjectChar("city")
    address = KeyObjectComposition("address",[country,city])
    person = KeyObjectComposition("person",[first_name,last_name,yob,emails,address])

    objs = [email_address
             , email_type
             , email
             , emails
             , first_name
             , last_name
             , yob
             , country
             , city
             , address
             , person]

    # KeySqlStore: __init__(self, name: str, catalog_name: str, schema_name="", owner_name="", host_names=[], hosts=[], component_names=[], constraints=[]):
    people_store = KeySqlStore("people_store", catalog_name=catalog_name)

    stmts = [people.create(), people.create_store(people_store)]

    for obj in objs:
        stmts.append(people.create_object(obj))

    objects = {obj.get_name() : obj for obj in objs}

    return people, people_store, objects, stmts


def drop(store : KeySqlStore,catalog : KeySqlCatalog):
    return [store.drop(),catalog.drop()]

if __name__ == "__main__":

    logging.basicConfig(level = logging.INFO)

    parser = argparse.ArgumentParser()

    # Host can be ip address or full_name, examples:  "10.0.0.123", "my_server_name.com"
    parser.add_argument('--host', required=False, default='ip_address_or_host_name.com', type=str)
    parser.add_argument('--port', required=False, default=1234, type=int)
    parser.add_argument('--user', required=True, type=str)
    parser.add_argument('--password', required=True, type=str)

    args = parser.parse_args()

    logging.info('------------------------------')
    # instantiate client
    keysql = KeySqlClient(args.host, args.port)
    conn = None
    try:
        # connect with server
        conn = keysql.connect(args.user, args.password)
        catalog, store, objects, statements = create_schema()

        # drop store if exists
        try:
            conn.execute(store.drop())
        except:
            pass
        # drop catalog if exists
        try:
            conn.execute(catalog.drop())
        except:
            pass

        # connection accepts list of statements
        conn.execute(statements)

        # insert accepts python classes, or dictionaries, or mix of both as input
        # insert returns a number of inserted instances
        count = conn.execute(store.insert([
                          Person(last_name='Johns', yob=1999, emails=[Email(email_address='shopping@johns.com'),
                                                                      Email(email_address='amazon.shopping@johns.com')]),
                          {"Person":
                               {"first_name": 'John',
                                "last_name": 'Smith',
                                "yob": 1997,
                                "address": {"country": "US", "city": "San francisco"},
                                "emails": [Email(email_address = "smith@work.org", email_type="work"),
                                           Email(email_address = "John@smith.org", email_type="home")]
                                }
                           }]))
        assert count == 2

        # verify that store has object PERSON
        # store - keysql format
        # {STORE_DESCRIPTION:{SCHEMA_NAME:'TEST3',CATALOG_NAME:'PEOPLE',STORE_NAME:'PEOPLE_STORE',OWNER_NAME:'test3',
        # HOSTS:{KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'PERSON',KEYOBJECT_TYPE:'COMPOSITION',
        # COMPONENT_KEYOBJECT_NAMES:{KEYOBJECT_NAME:'EMAILS',KEYOBJECT_NAME:'FIRST_NAME',KEYOBJECT_NAME:'LAST_NAME',
        # KEYOBJECT_NAME:'YOB',KEYOBJECT_NAME:'ADDRESS'}}},CONSTRAINTS:NULL}}

        store_instance = conn.execute(store.show())
        store_class = KeySqlClassFromInstance.create_store_from_instance(store_instance[0])
        #catalog_name, hosts_objs = conn.execute(store.show())
        catalog_name = store_class._catalog_name
        hosts_objs = store_class._host_names

        assert catalog_name == catalog.get_name().upper()
        assert len(hosts_objs) == 1 and hosts_objs[0] == 'PERSON'

        # select PERSON as python closs
        people = conn.execute(store.select(Person()))
        assert len(people) == 2

        # select multiple object names as strings
        emails = conn.execute(store.select("last_name","email").where("first_name = 'John'"))
        # {RESULT:{LAST_NAME:'Smith',#EMAIL:{EMAIL:{EMAIL_ADDRESS:'smith@work.org',EMAIL_TYPE:'work'},EMAIL:{EMAIL_ADDRESS:'John@smith.org',EMAIL_TYPE:'home'}}}}
        #email_class = KeySqlClassFromInstance.create_keyobject_from_instance(emails[0])
        #print(email_class, '------------------------------email_class')
        assert len(emails) == 1
        assert emails[0][0] == 'Smith'
        assert emails[0][1][0].EMAIL_TYPE in ['work','home']

        # select with group by and order by clauses
        emails = conn.execute(store.select("last_name","email").where("first_name = 'John'").group_by("email").order_by("email_type"))
        assert len(emails) == 2
        assert emails[0][1].EMAIL_TYPE == 'home'

        # select results can be dictionaries
        people = conn.execute(store.select(Person()).order_by("last_name").limit(1).into_dict())
        assert len(people) == 1
        assert people[0].LAST_NAME == 'Johns'
        assert people[0].FIRST_NAME is None

        addresses = conn.execute(store.select("last_name","address").into_dict())
        assert len(addresses) == 2

        # Create object for one select and then drop it
        person_name = KeyObjectComposition("person_name",[objects["last_name"],objects["first_name"]])
        assert person_name.get_name() == "person_name"
        conn.execute(catalog.create_object(person_name))
        assert person_name.get_name() == "person_name"

        # AS is implemented with dictionary
        person_names = conn.execute(store.select({"person_name" : ["last_name","first_name"]}))
        assert len(addresses) == 2
        conn.execute(catalog.drop_object(person_name))

        # aggregation function
        cnt = KeyObjectNumber("cnt")
        conn.execute(catalog.create_object(cnt))
        count = conn.execute(store.select({"cnt": "count(*)"}))
        assert len(count) == 1
        assert count[0] == 2

        count = conn.execute(store.select({"last_name": "last_name", "cnt": "count(email)"}).group_by("last_name").having("cnt > 1"))
        assert len(count) == 2
        assert count[0][1] == 2

        conn.execute(catalog.drop_object(cnt))

        # set in update parameters is dictionaries
        count = conn.execute(store.update().where("last_name = 'Johns'").set({"address" : {"country" : "Russia"}}))
        assert count == 1

        # update returns a number of updated rows
        count = conn.execute(store.update().where("last_name = 'Johns'").set({"address.city": "Moscow"}))
        assert count == 1

        # delete return a number of deleted rows
        count = conn.execute(store.delete().where("last_name = 'Johns'"))
        assert count == 1

        people = conn.execute(store.select(Person()))
        assert len(people) == 1

        # drop store, catalog
        conn.execute(store.drop())
        conn.execute(KeySqlDropCatalogStatement("people"))

        logging.info('OK')
    except ValueError as ex:
        logging.error(ex)   # ex.message())
    finally:
        if conn:
            conn.disconnect()