#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/

import sys
sys.path.append("..")
sys.path.append(".")
import logging
import argparse
import json
from pathlib import Path
from keysql_client.client import KeySqlClient
from keysql_client.key_object import KEYOBJECT_TYPE, KeyObjectChar, KeyObjectNumber,KeyObjectComposition, KeyObjectMultiple
from keysql_client.key_object_statement import  ShowKeyObjectStatement, CreateKeyObjectStatement
from keysql_client.keysql_store import KeySqlStore
from keysql_client.keysql_store_statement import KeySqlShowStoresStatement
from keysql_client.keysql_catalog import KeySqlCatalog, KeySqlShowCatalogStatement, KeySqlDropCatalogStatement, KeySqlListCatalogsStatement
from keysql_client.keysql_class_from_instance import KeySqlClassFromInstance

from typing import List
import locale
#locale.setlocale(locale.LC_ALL, 'en_US.utf-8')

# KeySql example (string equivalent)
catalog_schema = """
CREATE CATALOG MY_CATALOG_EXAMPLE;
CREATE KEYOBJECT AGE NUMBER IN CATALOG MY_CATALOG_EXAMPLE;
CREATE KEYOBJECT LAST_NAME CHAR IN CATALOG MY_CATALOG_EXAMPLE;
CREATE KEYOBJECT FIRST_NAME CHAR IN CATALOG MY_CATALOG_EXAMPLE;
CREATE KEYOBJECT PERSON {FIRST_NAME,LAST_NAME, AGE} IN CATALOG MY_CATALOG_EXAMPLE;
CREATE STORE MY_STORE_EXAMPLE FOR CATALOG MY_CATALOG_EXAMPLE;
"""

my_store_instances = """
INSERT INTO MY_STORE_EXAMPLE INSTANCES
{PERSON:{FIRST_NAME:'Jane', LAST_NAME:'Doe', AGE:21}},
{PERSON:{FIRST_NAME:'John', LAST_NAME:'Smith', AGE:32}},
{PERSON:{FIRST_NAME:'Jane', LAST_NAME:'Brown', AGE:43}};
"""

# To perform create catalog, store and insert instances without using python driver classes:
# test_connection.execute(catalog_schema)
# test_connection.execute(my_store_instances)

# help functions ============================================================================================
def write_to_file(path, str_data):
    with open(path, 'w') as f:
        for s in str_data:
            f.write(s)


def get_json_instances_from_file(file_name: str, host_name=None):
    """
    Most important function - if we can read json file as dictionary, we can insert
    instances in the form of dictionaries into store.
    :param file_name: file name with instances in json format
    :param host_name: Host name (string)
    :return:  list of dictionaries with instances
    """
    with open(file_name, 'r') as ff:
        instances = json.load(ff)
    if len(instances) > 0 and host_name:
        instances = [{host_name: item} for item in instances]
    return instances

# ============================================================================================

# temp file tmp2.json   - example with host included  (PERSON)
set2_with_host = """
[{"PERSON": {"FIRST_NAME": "Jane2", "LAST_NAME": "Doe2", "AGE": 21}},
{"PERSON": {"FIRST_NAME": "John2", "LAST_NAME": "Smith2", "AGE": 32}},
{"PERSON": {"FIRST_NAME": "Jane2", "LAST_NAME": "Brown2", "AGE": 43}}]
"""
path = Path(__file__).parent.absolute()  # folder with this file
path_set2 = "{}/tmp2.json".format(path)
write_to_file(path_set2, set2_with_host)

# temp file tmp3.json   - example with host not included => to insert, host needed to transfer into form tmp2.json
# PERSON is the host for this file
set3_without_host = """
[{"FIRST_NAME":"Jane3","LAST_NAME":"Doe3","AGE":21},
{"FIRST_NAME":"John3","LAST_NAME":"Smith3","AGE":null},
{"FIRST_NAME":"Jane3","LAST_NAME":null,"AGE":43},
{"LAST_NAME":"Doe3","AGE":21}]
"""
path_set3 = "{}/tmp3.json".format(path)
write_to_file(path_set3, set3_without_host)

def clean_catalog_for_connection(connection, catalog_name, schema=None, all=False) -> None:
    """
    Clean catalog for connection. If it does not exists do nothing
    :param connection:
    :param catalog_name: can be full name or not, sch.name; if sch != schema (for non-empty) => error
    :return:
    """
    # use default schema
    if schema is None:
        schema = connection.schema.upper()

    list_catalogs_statement = "list catalogs ALL"
    all_catalog_names = connection.execute(list_catalogs_statement)

    catalog_name = catalog_name.upper()
    if '.' not in catalog_name:
        catalog_name = schema + '.' + catalog_name

    if catalog_name not in all_catalog_names:
        return

    stores = connection.execute('list stores for catalog {}'.format(catalog_name)) # , format='json')
    # if there are stores, remove them
    for _store in stores:
        reply = connection.execute('drop store {}.{}'.format(schema, _store))

    # we delete catalog
    connection.execute("drop catalog {}".format(catalog_name))


def create_catalog():
     # drop catalog if exists
    if CLEAN_CATALOG:
        clean_catalog_for_connection(test_connection, catalog_name, schema=schema, all=False)

    logging.info('step 1: create catalog if it does not exist ---------------------------')
    # if catalog is not created from the scratch, some/all keyobjects can be already created,
    # check all before creating
    all_catalogs = test_connection.execute(KeySqlListCatalogsStatement(schema=schema))
    create_catalog_from_scratch = catalog_name not in all_catalogs

    if create_catalog_from_scratch:
        test_connection.execute(ko_catalog.create())

    return create_catalog_from_scratch

def create_keyobjects():
    reply = test_connection.execute(KeySqlShowCatalogStatement(catalog_name))
    reply_catalog_class = KeySqlClassFromInstance.create_catalog_from_instance(reply[0])
    created_already_keyobjects = reply_catalog_class.get_keyobject_names()

    reply = test_connection.execute(KeySqlShowCatalogStatement(catalog_name))
    reply_catalog_class = KeySqlClassFromInstance.create_catalog_from_instance(reply[0])
    created_already_keyobjects = reply_catalog_class.get_keyobject_names()

    logging.info('step 2: create objects---------------------------')
    # create keyobjects if they are not in catalog
    # First_name
    if ko_first_name.get_name() not in created_already_keyobjects:
        test_connection.execute(CreateKeyObjectStatement(ko_first_name, catalog_name))
    reply_first_name = test_connection.execute(ShowKeyObjectStatement(ko_first_name.get_name(), catalog_name))
    first_name_ko_class = KeySqlClassFromInstance.create_keyobject_from_instance(reply_first_name)
    assert first_name_ko_class.get_name() == ko_first_name.get_name().upper()
    assert first_name_ko_class.keyobject_type() == KEYOBJECT_TYPE.CHAR

    # Last_name
    if ko_last_name.get_name() not in created_already_keyobjects:
        test_connection.execute(CreateKeyObjectStatement(ko_last_name, catalog_name))
    reply_last_name = test_connection.execute(ShowKeyObjectStatement(ko_last_name.get_name(), catalog_name))
    last_name_ko_class = KeySqlClassFromInstance.create_keyobject_from_instance(reply_last_name)
    assert last_name_ko_class.get_name() == ko_last_name.get_name().upper()
    assert last_name_ko_class.keyobject_type() == KEYOBJECT_TYPE.CHAR

    # Age
    if ko_age.get_name() not in created_already_keyobjects:
        test_connection.execute(CreateKeyObjectStatement(ko_age, catalog_name))
    reply_age = test_connection.execute(ShowKeyObjectStatement(ko_age.get_name(), catalog_name))
    age_ko_class = KeySqlClassFromInstance.create_keyobject_from_instance(reply_age)
    assert age_ko_class.get_name() == ko_age.get_name().upper()
    assert age_ko_class.keyobject_type() == KEYOBJECT_TYPE.NUMBER

    # Person
    if ko_person.get_name() not in created_already_keyobjects:
        test_connection.execute(CreateKeyObjectStatement(ko_person, catalog_name))

    reply = test_connection.execute(ShowKeyObjectStatement(ko_person.get_name(), catalog_name))
    ko_classes = {
        "FIRST_NAME": first_name_ko_class,
        "LAST_NAME": last_name_ko_class,
        "AGE": age_ko_class}
    person_ko_class = KeySqlClassFromInstance.create_keyobject_from_instance(reply, ko_classes=ko_classes)
    assert person_ko_class.get_name() == ko_person.get_name().upper()
    assert person_ko_class.keyobject_type() == KEYOBJECT_TYPE.COMPOSITION
    assert len(person_ko_class.get_children()) == 3
    for x in person_ko_class.get_children():
        assert x.get_name().lower() in ["first_name", "last_name", "age"]

    logging.info('step 3: check objects---------------------------')
    reply = test_connection.execute(KeySqlShowCatalogStatement(catalog_name))
    catalog_class = KeySqlClassFromInstance.create_catalog_from_instance(reply[0])
    assert len(catalog_class.get_keyobject_names()) == 4
    for x in catalog_class.get_keyobject_names():
        assert x.lower() in ["first_name", "last_name", "age", "person"]

    return created_already_keyobjects

def create_store():
    logging.info('step 4: create store -----------------------------------------------------------------')
    all_stores = test_connection.execute(KeySqlShowStoresStatement(catalog_name))  # all stores in catalog
    if store_name not in all_stores:
        test_connection.execute(ko_catalog.create_store(ko_store))

    logging.info('step 5: insert instances as dictionaries  --------------------------------------------')
    count = test_connection.execute(ko_store.insert([
        {"PERSON": {"FIRST_NAME": 'Jane', "LAST_NAME": 'Doe', "AGE": 21}},
        {"PERSON": {"FIRST_NAME": 'John', "LAST_NAME": 'Smith', "AGE": 32}},
        {"PERSON": {"FIRST_NAME": 'Jane', "LAST_NAME": 'Brown', "AGE": 43}}
    ]))

    assert count == 3

    # verify that catalog and store are created and have data
    reply = test_connection.execute("show catalog {}".format(catalog_name), format='json', keep_str_res=True)
    logging.info('catalog in json format:----------------------------------------------------------------')
    logging.info(test_connection.get_last_str_result())

    reply = test_connection.execute("show catalog {}".format(catalog_name), keep_str_res=True)
    logging.info('catalog: in keysql format: -------------------------------------------------------------------')
    logging.info(test_connection.get_last_str_result())


# Instructions:
# if python driver not in the path run from the folder where keysql_client  is installed
# run script from command line or from pycharm's therminal
# python example_read_instances_json_format.py  --host=<your_host> --port=<your_port> --user=<user_name> --password=<your_pass> --schema=<your_schema> --clean=<True/False>
# Host can be ip address or full_name, examples:  "10.0.0.123", "my_server_name.com"
#  --clean: if --clean=True - create catalog from scratch  (delete catalog if existed)
#           if --clean=False - not to delete catalog, keyobjects, and instances
#           default:  --clean=False

if __name__ == "__main__":

    logging.basicConfig(level = logging.INFO)

    # if run from command line
    parser = argparse.ArgumentParser()

    parser.add_argument('--host', required=True, type=str)
    parser.add_argument('--port', required=True, type=int)
    parser.add_argument('--user', required=True, type=str)
    parser.add_argument('--password', required=True, type=str)
    parser.add_argument('--schema', required=True, type=str)
    parser.add_argument('--clean', required=False, default=False, type=str)

    args = parser.parse_args()
    user = args.user
    password = args.password
    host = args.host
    port = args.port
    schema = args.schema
    CLEAN_CATALOG = args.clean

    logging.info('---------- start -----------------------------------------------------------------------')
    # instatiate client
    keysql = KeySqlClient(host, port)

    catalog_name = "MY_CATALOG_EXAMPLE"
    store_name = "MY_STORE_EXAMPLE"
    test_connection = None # if try:... block fails
    try:
        # connect with server
        test_connection = keysql.connect(user, password, schema, True)

        ko_first_name = KeyObjectChar('FIRST_NAME')
        ko_last_name = KeyObjectChar('LAST_NAME')
        ko_age = KeyObjectNumber('AGE')
        ko_person = KeyObjectComposition('PERSON', [ko_first_name, ko_last_name, ko_age])

        ko_catalog = KeySqlCatalog(catalog_name)
        ko_store = KeySqlStore(store_name, catalog_name)

        create_catalog_from_scratch = create_catalog()
        created_already_keyobjects = create_keyobjects()
        create_store()

        logging.info('read json file with host objects.......................................................')
        json_res_from_file = get_json_instances_from_file(path_set2)

        # check inserted instances
        reply_keysql = test_connection.execute("select * from {}".format(store_name), format='keysql', keep_str_res=True)
        logging.info('select statement: (keysql instance)------------------------------------------------------')
        logging.info(test_connection.get_last_str_result())

        logging.info('step 5: insert instances from json file -------------------------------------------------')
        count = test_connection.execute(ko_store.insert(json_res_from_file))
        assert count == 3

        # check inserted instances
        reply_keysql = test_connection.execute("select * from {}".format(store_name), format='keysql', keep_str_res=True)
        # result should be  (if created from scratch)
        # {PERSON:{FIRST_NAME:'Jane',LAST_NAME:'Doe',AGE:21}},
        # {PERSON:{FIRST_NAME:'John',LAST_NAME:'Smith',AGE:32}},
        # {PERSON:{FIRST_NAME:'Jane',LAST_NAME:'Brown',AGE:43}},
        # {PERSON:{FIRST_NAME:'Jane2',LAST_NAME:'Doe2',AGE:21}},
        # {PERSON:{FIRST_NAME:'John2',LAST_NAME:'Smith2',AGE:32}},
        # {PERSON:{FIRST_NAME:'Jane2',LAST_NAME:'Brown2',AGE:43}}
        if create_catalog_from_scratch:
            assert len(reply_keysql) == 6
        logging.info(test_connection.get_last_str_result())

        logging.info('step 6: read json file without host objects...............................................')
        tmp_file_name = "../keysql_client/tmp3.json"

        host_name = "PERSON"
        json_res_from_file_add_host = get_json_instances_from_file(path_set3, host_name=host_name)
        count = test_connection.execute(ko_store.insert(json_res_from_file_add_host))
        assert count == 4

        # check all instances
        reply_keysql = test_connection.execute("select * from {}".format(store_name), format='keysql', keep_str_res=True)
        # Final result should be: (if created from scratch)
        # {PERSON:{FIRST_NAME:'Jane',LAST_NAME:'Doe',AGE:21}},
        # {PERSON:{FIRST_NAME:'John',LAST_NAME:'Smith',AGE:32}},
        # {PERSON:{FIRST_NAME:'Jane',LAST_NAME:'Brown',AGE:43}},
        # {PERSON:{FIRST_NAME:'Jane2',LAST_NAME:'Doe2',AGE:21}},
        # {PERSON:{FIRST_NAME:'John2',LAST_NAME:'Smith2',AGE:32}},
        # {PERSON:{FIRST_NAME:'Jane2',LAST_NAME:'Brown2',AGE:43}},
        # {PERSON:{FIRST_NAME:'Jane3',LAST_NAME:'Doe3',AGE:21}},
        # {PERSON:{FIRST_NAME:'John3',LAST_NAME:'Smith3',AGE:NULL}},
        # {PERSON:{FIRST_NAME:'Jane3',LAST_NAME:NULL,AGE:43}},
        # {PERSON:{FIRST_NAME:NULL,LAST_NAME:'Doe3',AGE:21}}

        if create_catalog_from_scratch:
            assert len(reply_keysql) == 10
        logging.info(test_connection.get_last_str_result())

        # clean catalog again
        if CLEAN_CATALOG:
            clean_catalog_for_connection(test_connection, catalog_name, schema=schema, all=False)
        if test_connection.is_connected():
            test_connection.disconnect()
        logging.info('OK................................................................................')

    except ValueError as ex:
        logging.error(ex)   # ex.message())
    finally:
        if test_connection and test_connection.is_connected():
            # in case catalog is not deleted
            if CLEAN_CATALOG:
                clean_catalog_for_connection(test_connection, catalog_name, schema=schema, all=False)
            test_connection.disconnect()
