#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/

import sys
sys.path.append("..")
sys.path.append(".")
import logging
import argparse
import json
from pathlib import Path
from keysql_client.client import KeySqlClient
from keysql_client.key_object import KEYOBJECT_TYPE, KeyObjectChar, KeyObjectNumber,KeyObjectComposition, KeyObjectMultiple
from keysql_client.key_object_statement import  ShowKeyObjectStatement, CreateKeyObjectStatement
from keysql_client.keysql_store import KeySqlStore
from keysql_client.keysql_store_statement import KeySqlShowStoresStatement
from keysql_client.keysql_catalog import KeySqlCatalog, KeySqlShowCatalogStatement, KeySqlDropCatalogStatement, KeySqlListCatalogsStatement
from keysql_client.keysql_class_from_instance import KeySqlClassFromInstance

from typing import List
import locale
#locale.setlocale(locale.LC_ALL, 'en_US.utf-8')

# KeySql example (string equivalent)
catalog_schema = """
CREATE CATALOG MY_CATALOG_EXAMPLE;
CREATE KEYOBJECT AGE NUMBER IN CATALOG MY_CATALOG_EXAMPLE;
CREATE KEYOBJECT LAST_NAME CHAR IN CATALOG MY_CATALOG_EXAMPLE;
CREATE KEYOBJECT FIRST_NAME CHAR IN CATALOG MY_CATALOG_EXAMPLE;
CREATE KEYOBJECT PERSON {FIRST_NAME,LAST_NAME, AGE} IN CATALOG MY_CATALOG_EXAMPLE;
CREATE STORE MY_STORE_EXAMPLE FOR CATALOG MY_CATALOG_EXAMPLE;
"""

my_store_instances = """
INSERT INTO MY_STORE_EXAMPLE INSTANCES
{PERSON:{FIRST_NAME:'Jane', LAST_NAME:'Doe', AGE:21}},
{PERSON:{FIRST_NAME:'John', LAST_NAME:'Smith', AGE:32}},
{PERSON:{FIRST_NAME:'Jane', LAST_NAME:'Brown', AGE:43}};
"""
# To perform create catalog, store and insert instances without using python driver classes:
# test_connection.execute(catalog_schema)
# test_connection.execute(my_store_instances)

""" 
This version does not check if catalog is created (to simpify code and remove all checks:
if stores, keyobjects are created.
The catalog should be new !!!
"""

# help functions ============================================================================================
def write_to_file(path, str_data):
    """
    Write string data (list of strings) to file
    :param path:
    :param str_data:
    :return:
    """
    with open(path, 'w') as f:
        for s in str_data:
            f.write(s)

def get_json_instances_from_file(file_name: str, host_name=None):
    """
    Most important function - if we can read json file as dictionary, we can insert
    instances in the form of dictionaries into store.
    :param file_name: file name with instances in json format
    :param host_name: Host name (string)
    :return:  list of dictionaries with instances
    """
    with open(file_name, 'r') as ff:
        instances = json.load(ff)
    if len(instances) > 0 and host_name:
        instances = [{host_name: item} for item in instances]
    return instances

# ============================================================================================
# Prepare json data
# Write two file with json format (I put it here to show the view of the files)
# temp file tmp2.json   - example with host included  (PERSON)
set2_with_host = """
[{"PERSON": {"FIRST_NAME": "Jane2", "LAST_NAME": "Doe2", "AGE": 21}},
{"PERSON": {"FIRST_NAME": "John2", "LAST_NAME": "Smith2", "AGE": 32}},
{"PERSON": {"FIRST_NAME": "Jane2", "LAST_NAME": "Brown2", "AGE": 43}}]
"""
path = Path(__file__).parent.absolute()  # folder with this file
path_set2 = "{}/tmp2.json".format(path)
write_to_file(path_set2, set2_with_host)

# temp file tmp3.json   - example with host not included => to insert, host needed to transfer into form tmp2.json
# Some attributes are missing - they will filled with NULL (default values)
# PERSON is the host for this file
set3_without_host = """
[{"FIRST_NAME":"Jane3","LAST_NAME":"Doe3","AGE":21},
{"FIRST_NAME":"John3","LAST_NAME":"Smith3","AGE":null},
{"FIRST_NAME":"Jane3","LAST_NAME":null,"AGE":43},
{"LAST_NAME":"Doe3","AGE":21}]
"""
path_set3 = "{}/tmp3.json".format(path)
write_to_file(path_set3, set3_without_host)
# ============================================================================================
# steps in pipeline

def create_catalog():
    """
    We do not check if catalog or store exist !!! Catalog and store should be new (for the schema)!!!!
    Catalog name and store name should be changed in the main function                            !!!!
    """
    logging.info('step 1: create catalog if it does not exist ---------------------------')
    test_connection.execute(ko_catalog.create())


def create_keyobjects():
    """
    We create new catalog, so it is empty and there is no need to check if objects exist
    """
    logging.info('step 2: create objects---------------------------')
    # create keyobjects
    # First_name
    test_connection.execute(CreateKeyObjectStatement(ko_first_name, catalog_name))

    # Last_name
    test_connection.execute(CreateKeyObjectStatement(ko_last_name, catalog_name))

    # Age
    test_connection.execute(CreateKeyObjectStatement(ko_age, catalog_name))

    # Person
    test_connection.execute(CreateKeyObjectStatement(ko_person, catalog_name))


def create_store():
    test_connection.execute(ko_catalog.create_store(ko_store))

    logging.info('step 5: insert instances as dictionaries  --------------------------------------------')
    count = test_connection.execute(ko_store.insert([
        {"PERSON": {"FIRST_NAME": 'Jane', "LAST_NAME": 'Doe', "AGE": 21}},
        {"PERSON": {"FIRST_NAME": 'John', "LAST_NAME": 'Smith', "AGE": 32}},
        {"PERSON": {"FIRST_NAME": 'Jane', "LAST_NAME": 'Brown', "AGE": 43}}
    ]))
    assert count == 3
# ============================================================================================

# Instructions:
# if python driver not in the path run from the folder where keysql_client  is installed
# run script from command line or from pycharm's therminal
# python example_read_instances_json_format.py  --host=<your_host> --port=<your_port> --user=<user_name> --password=<your_pass> --schema=<your_schema>

if __name__ == "__main__":

    logging.basicConfig(level = logging.INFO)
    parser = argparse.ArgumentParser()

    parser.add_argument('--host', required=True, type=str)
    parser.add_argument('--port', required=True, type=int)
    parser.add_argument('--user', required=True, type=str)
    parser.add_argument('--password', required=True, type=str)
    parser.add_argument('--schema', required=True, type=str)

    args = parser.parse_args()
    user = args.user
    password = args.password
    host = args.host
    port = args.port
    schema = args.schema

    logging.info('---------- start -----------------------------------------------------------------------')
    # instatiate client
    keysql = KeySqlClient(host, port)

    # We do not check if catalog or store exist !!! Catalog and store should be new (for the schema)!!!!
    # Catalog name and store name should be changed in the main function                            !!!!

    catalog_name = "MY_CATALOG_EXAMPLE5"
    store_name = "MY_STORE_EXAMPLE5"
    #-------------------------------------------------------------------------------------------------------

    test_connection = None # if try:... block fails
    try:
        # connect with server
        test_connection = keysql.connect(user, password, schema, True)

        ko_first_name = KeyObjectChar('FIRST_NAME')
        ko_last_name = KeyObjectChar('LAST_NAME')
        ko_age = KeyObjectNumber('AGE')
        ko_person = KeyObjectComposition('PERSON', [ko_first_name, ko_last_name, ko_age])

        ko_catalog = KeySqlCatalog(catalog_name)
        ko_store = KeySqlStore(store_name, catalog_name)

        create_catalog()
        create_keyobjects()
        create_store()

        logging.info('read json file with host objects.......................................................')
        json_res_from_file = get_json_instances_from_file(path_set2)

        # check inserted instances
        reply_keysql = test_connection.execute("select * from {}".format(store_name), format='keysql', keep_str_res=True)
        logging.info('select statement: (keysql instance)------------------------------------------------------')
        logging.info(test_connection.get_last_str_result())

        logging.info('step 5: insert instances from json file -------------------------------------------------')
        count = test_connection.execute(ko_store.insert(json_res_from_file))
        assert count == 3

        # check inserted instances
        reply_keysql = test_connection.execute("select * from {}".format(store_name), format='keysql', keep_str_res=True)
        # result should be  (if created from scratch)
        # {PERSON:{FIRST_NAME:'Jane',LAST_NAME:'Doe',AGE:21}},
        # {PERSON:{FIRST_NAME:'John',LAST_NAME:'Smith',AGE:32}},
        # {PERSON:{FIRST_NAME:'Jane',LAST_NAME:'Brown',AGE:43}},
        # {PERSON:{FIRST_NAME:'Jane2',LAST_NAME:'Doe2',AGE:21}},
        # {PERSON:{FIRST_NAME:'John2',LAST_NAME:'Smith2',AGE:32}},
        # {PERSON:{FIRST_NAME:'Jane2',LAST_NAME:'Brown2',AGE:43}}
        assert len(reply_keysql) == 6
        logging.info('after adding instances with host')
        logging.info(test_connection.get_last_str_result())

        logging.info('step 6: read json file without host objects...............................................')
        tmp_file_name = "../keysql_client/tmp3.json"

        host_name = "PERSON"
        json_res_from_file_add_host = get_json_instances_from_file(path_set3, host_name=host_name)
        count = test_connection.execute(ko_store.insert(json_res_from_file_add_host))
        assert count == 4

        # check all instances
        reply_keysql = test_connection.execute("select * from {}".format(store_name), format='keysql', keep_str_res=True)
        # Final result should be: (if created from scratch)
        # {PERSON:{FIRST_NAME:'Jane',LAST_NAME:'Doe',AGE:21}},
        # {PERSON:{FIRST_NAME:'John',LAST_NAME:'Smith',AGE:32}},
        # {PERSON:{FIRST_NAME:'Jane',LAST_NAME:'Brown',AGE:43}},
        # {PERSON:{FIRST_NAME:'Jane2',LAST_NAME:'Doe2',AGE:21}},
        # {PERSON:{FIRST_NAME:'John2',LAST_NAME:'Smith2',AGE:32}},
        # {PERSON:{FIRST_NAME:'Jane2',LAST_NAME:'Brown2',AGE:43}},
        # {PERSON:{FIRST_NAME:'Jane3',LAST_NAME:'Doe3',AGE:21}},
        # {PERSON:{FIRST_NAME:'John3',LAST_NAME:'Smith3',AGE:NULL}},
        # {PERSON:{FIRST_NAME:'Jane3',LAST_NAME:NULL,AGE:43}},
        # {PERSON:{FIRST_NAME:NULL,LAST_NAME:'Doe3',AGE:21}}

        assert len(reply_keysql) == 10
        logging.info('after adding instances without host')
        logging.info(test_connection.get_last_str_result())

        # finish job - disconnect
        test_connection.disconnect()
        logging.info('OK................................................................................')

    except ValueError as ex:
        logging.error(ex)   # ex.message())
    finally:
        if test_connection and test_connection.is_connected():
            # in case catalog is not deleted
            test_connection.disconnect()
